#include <sys/types.h>
#include <sys/stat.h>
#include <sys/termios.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>

#include <syslog.h>
#include <fcntl.h>
#include <sys/resource.h>

#include <sched.h>

#define MAXLINE 4096
const char* daemon_name = "rtsched";

static void err_doit(int errnoflag, int error, const char *fmt, va_list ap)
{
	char buf[MAXLINE];
	vsnprintf(buf, MAXLINE, fmt, ap);
	if (errnoflag)
		snprintf(buf+strlen(buf), MAXLINE-strlen(buf), ": %s",
	strerror(error));
	strcat(buf, "\n");
	fflush(stdout);
	fputs(buf, stderr);
	fflush(NULL);
}

void err_quit(const char *fmt, ...)
{
	va_list     ap;

	va_start(ap, fmt);
	err_doit(0, 0, fmt, ap);
	va_end(ap);
	exit(1);
}

void daemonize(const char *cmd)
{
	int               i;
	int               fd0, fd1, fd2;
	pid_t             pid;
	struct rlimit     rl;
	struct sigaction  sa;

	/* Clear file creation mask.  */
	umask(0);

	/* Get maximum number of file descriptors.  */
	if (getrlimit(RLIMIT_NOFILE, &rl) < 0)
		err_quit("%s: can't get file limit", cmd);

	/* Become a session leader to lose controlling TTY.  */
	if ((pid = fork()) < 0)
		err_quit("%s: can't fork", cmd);
	else if (pid != 0) /* parent */
		exit(0);
	setsid();

	/* Ensure future opens won't allocate controlling TTYs.  */
	sa.sa_handler = SIG_IGN;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	if (sigaction(SIGHUP, &sa, NULL) < 0)
		err_quit("%s: can't ignore SIGHUP");
	if ((pid = fork()) < 0)
		err_quit("%s: can't fork", cmd);
	else if (pid != 0) /* parent */
		exit(0);

	/* Change the current working directory to the root */
	if (chdir("/") < 0)
		err_quit("%s: can't change directory to /");

	/* Close all open file descriptors. */
	if (rl.rlim_max == RLIM_INFINITY)
		rl.rlim_max = 1024;
	for (i = 0; i < rl.rlim_max; i++)
		close(i);

	/* Attach file descriptors 0, 1, and 2 to /dev/null.  */
	fd0 = open("/dev/null", O_RDWR);
	fd1 = dup(0);
	fd2 = dup(0);

	/* Initialize the log file. */
	openlog(cmd, LOG_CONS, LOG_DAEMON);
	if (fd0 != 0 || fd1 != 1 || fd2 != 2) {
		syslog(LOG_ERR, "unexpected file descriptors %d %d %d",
			fd0, fd1, fd2);
		exit(1);
	}
}

int bind_to_cpu(int cpu_nb)
{
	cpu_set_t mask;
	__CPU_ZERO(&mask);
	__CPU_SET(cpu_nb,&mask);
	return sched_setaffinity(0, sizeof(cpu_set_t), &mask);
}

static __inline__ unsigned long long rdrtc(void)
{
	unsigned long long result=0;
	unsigned long int upper, lower, tmp;
	__asm__ volatile(
		"0:                  \n"
		"\tmftbu   %0           \n"
		"\tmftb    %1           \n"
		"\tmftbu   %2           \n"
		"\tcmpw    %2,%0        \n"
		"\tbne     0b         \n"
		: "=r"(upper),"=r"(lower),"=r"(tmp)
	);
	result = upper;
	result = (result<<32) | lower;
	return(result);
}

unsigned long long rtc_per_100ms = 0;

static unsigned long long get_rtc_per_100ms(void)
{
	unsigned long long prev_rtc;

	if(!rtc_per_100ms){
		prev_rtc = rdrtc();
		usleep(100*1000);
		rtc_per_100ms = rdrtc() - prev_rtc;	
	}
	return rtc_per_100ms;
}

#define BUFF_LEN (1024UL*6)

char big_buf[BUFF_LEN]={0};

int do_one_loop(void)
{
	int index;
	int result = 0;

	for (index = 0; index < BUFF_LEN; index ++)
		big_buf[index] = random();

	for (index =0; index < BUFF_LEN; index++)
		result = big_buf[index] * result;

	return result;
}

unsigned int measure_one_loop(void)
{
	unsigned int loops_per_100ms = 0;
	unsigned long long prev_rtc;

	prev_rtc = rdrtc();
	while(rdrtc() - prev_rtc < get_rtc_per_100ms()){
		do_one_loop();
		loops_per_100ms++;
	}
	return loops_per_100ms;
}

int main(int argc, char** argv)
{
	int loop_no = 0;
	int opt;
	int cpu = 0;
	int sched_policy;	
	struct sched_param sched_param;
	int sched_prio;
	unsigned long long prev_rtc;
	unsigned int run_period_ms, sleep_period_ms;
	unsigned int loops_per_100ms;

	run_period_ms = sleep_period_ms = 300;
	sched_policy = SCHED_FIFO;
	sched_prio = sched_get_priority_max(sched_policy);


	/* get options */
	while ((opt = getopt(argc, argv, "l:p:f:r:x:y:m")) != EOF) {
		switch (opt) {
		case 'p':
			cpu = strtol(optarg, NULL, 0);
			break;
		case 'f':
			sched_policy = SCHED_FIFO;
			sched_prio = strtol(optarg, NULL, 0);
			break;
		case 'r':
			sched_policy = SCHED_RR;
			sched_prio = strtol(optarg, NULL, 0);
			break;
		case 'x':
			run_period_ms = strtol(optarg, NULL, 0);
			break;
		case 'y':
			sleep_period_ms = strtol(optarg, NULL, 0);
			break;
		case 'm':
			/* get rtc count per 100ms, SCHED_FIFO makes it more accurate */
			sched_param.sched_priority=1;
			if(sched_setscheduler(0,SCHED_FIFO,&sched_param)<0)
				err_quit("Error while setting scheduling policy.\n");
			get_rtc_per_100ms();
			loops_per_100ms = measure_one_loop();
			printf("loops_per_100ms: %d\n",loops_per_100ms);
			return loops_per_100ms;
			break;
		case 'l':
			loops_per_100ms = strtol(optarg, NULL, 0);
			break;
		}
	}

	/* make myself a daemon */
	daemonize(daemon_name);

	/* bind to the designated CPU */
	if(bind_to_cpu(cpu)<0){
		syslog(LOG_INFO,"Error while binding to cpu %d\n",cpu);
		exit(-1);
	}

	/* set scheduling policy */
	sched_param.sched_priority = sched_prio;
	if(sched_setscheduler(0, sched_policy, &sched_param)<0){
		syslog(LOG_INFO, "Error while setting scheduling policy.\n");
		exit(-1);
	}

	syslog(LOG_INFO, "Instance[%d]started at CPU %d, with scheduling %s:%d\n", 
	       getpid(), cpu, sched_policy==SCHED_FIFO?"SCHED_FIFO":"SCHED_RR", sched_prio);

	/* do something */
	while(1){
		int loops;
		/* run some ms */
		for (loops = 0; loops < loops_per_100ms * run_period_ms / 100; loops++){
			do_one_loop();
		}
		usleep(sleep_period_ms*1000);
		if(++loop_no%10==0)
			syslog(LOG_INFO, "Instance[%d] looped %d.\n", getpid(), loop_no);
	}
}

